"""Provides a confirmation dialog."""

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Label

##############################################################################
# Local imports.
from ..tools import add_key


##############################################################################
class Confirm(ModalScreen[bool]):
    """A modal dialog for confirming things."""

    CSS = """
    Confirm {
        align: center middle;

        &> Vertical {
            padding: 1 2;
            height: auto;
            width: auto;
            max-width: 80vw;
            background: $surface;
            border: panel $error;
            border-title-color: $text;

            &> Horizontal {
                height: auto;
                width: 100%;
                align-horizontal: center;
            }
        }

        Label {
            width: auto;
            max-width: 70vw;
            padding-left: 1;
            padding-right: 1;
            margin-bottom: 1;
        }

        Button {
            margin-right: 1;
        }
    }
    """

    BINDINGS = [
        ("escape", "no"),
        ("f2", "yes"),
        ("left", "app.focus_previous"),
        ("right", "app.focus_next"),
    ]

    def __init__(
        self, title: str, question: str, yes_text: str = "Yes", no_text: str = "No"
    ) -> None:
        """Initialise the dialog.

        Args:
            title: The title for the dialog.
            question: The question to ask the user.
            yes_text: The text for the yes button.
            no_text: The text for the no button.
        """
        super().__init__()
        self._title = title
        """The title for the dialog."""
        self._question = question
        """The question to ask the user."""
        self._yes = yes_text
        """The text of the yes button."""
        self._no = no_text
        """The text of the no button."""

    def compose(self) -> ComposeResult:
        """Compose the layout of the dialog."""
        with Vertical() as dialog:
            dialog.border_title = self._title
            yield Label(self._question)
            with Horizontal():
                yield Button(add_key(self._no, "Esc", self), id="no")
                yield Button(add_key(self._yes, "F2", self), id="yes")

    @on(Button.Pressed, "#yes")
    def action_yes(self) -> None:
        """Send back the positive response."""
        self.dismiss(True)

    @on(Button.Pressed, "#no")
    def action_no(self) -> None:
        """Send back the negative response."""
        self.dismiss(False)


### confirm.py ends here



"""Provides the base class for all application command messages."""

##############################################################################
# Backward compatibility.
from __future__ import annotations

##############################################################################
# Python imports.
from re import Pattern, compile
from typing import Final

##############################################################################
# Textual imports.
from textual.binding import Binding, BindingType
from textual.message import Message


##############################################################################
class Command(Message):
    """Base class for all application command messages."""

    COMMAND: str | None = None
    """The text for the command.

    Notes:
        If no `COMMAND` is provided the class name will be used.
    """

    ACTION: str | None = None
    """The action to call when the command is executed.

    By default the action will be:

       `action_{snake-case-of-command}_command`

    There are some commands that will result in unhelpful snake-case names,
    and also times where you want the action method named in a better way;
    in such cases use `ACTION` to override the method name.
    """

    FOOTER_TEXT: str | None = None
    """The text to show in the footer.

    Notes:
        If no `FOOTER_TEXT` is provided the
        [`command`][textual_enhanced.commands.Command.command] will be used.
    """

    SHOW_IN_FOOTER: bool = False
    """Should the command be shown in the [footer][textual.widgets.Footer]?

    By default commands are *not* shown in the
    [footer][textual.widgets.Footer], so set this to [`True`][True] to have
    the command appear there.
    """

    BINDING_KEY: str | tuple[str, str] | None = None
    """The binding key for the command.

    This can either be a string, which is the keys to bind, a tuple of the
    keys and also an overriding display value, or [`None`][None].
    """

    _SPLITTER: Final[Pattern[str]] = compile("[A-Z]?[a-z]+|[A-Z]+(?=[A-Z]|$)")
    """Regular expression for splitting up a command name."""

    @classmethod
    def command(cls) -> str:
        """The text for the command.

        Returns:
            The command's textual name.
        """
        return cls.COMMAND or " ".join(cls._SPLITTER.findall(cls.__name__))

    @property
    def context_command(self) -> str:
        """The command in context."""
        return self.command()

    @classmethod
    def tooltip(cls) -> str:
        """The tooltip for the command."""
        return cls.__doc__ or ""

    @property
    def context_tooltip(self) -> str:
        """The tooltip for the command, in context."""
        return self.tooltip()

    @classmethod
    def key_binding(cls) -> str:
        """Get the key that is the binding for this command.

        Returns:
            The key that is bound, or [`None`][None] if there isn't one.

        Notes:
            If a command has multiple bindings, only the first key is
            returned.
        """
        if isinstance(key := cls.BINDING_KEY, tuple):
            key, *_ = key
        if isinstance(key, str):
            key, *_ = cls.binding().key.partition(",")
        return (key or "").strip()

    @classmethod
    def action_name(cls) -> str:
        """Get the action name for the command.

        Returns:
            The action name.
        """
        return (
            cls.ACTION
            or f"{'_'.join(cls._SPLITTER.findall(cls.__name__))}_command".lower()
        )

    @staticmethod
    def bindings(*bindings: BindingType | type[Command]) -> list[BindingType]:
        """Create bindings.

        Args:
            bindings: Normal Textual bindings or a command class.

        Returns:
            A list of bindings that can be used with [`BINDINGS`][textual.dom.DOMNode.BINDINGS].
        """
        return [
            (binding if isinstance(binding, (Binding, tuple)) else binding.binding())
            for binding in bindings
        ]

    @classmethod
    def binding(cls) -> Binding:
        """Create a binding object for the command.

        Returns:
            A [`Binding`][textual.binding.Binding] for the command's key bindings.

        Raises:
            ValueError: If the command has no key binding.
        """
        if not cls.BINDING_KEY:
            raise ValueError("No binding key defined, unable to create a binding")
        keys, display = (
            cls.BINDING_KEY
            if isinstance(cls.BINDING_KEY, tuple)
            else (cls.BINDING_KEY, None)
        )
        return Binding(
            keys,
            cls.action_name(),
            description=cls.FOOTER_TEXT or cls.command(),
            tooltip=cls.tooltip(),
            show=cls.SHOW_IN_FOOTER,
            key_display=display,
            id=cls.__name__,
        )

    @property
    def has_binding(self) -> bool:
        """Does this command have a binding?"""
        return self.BINDING_KEY is not None


### command.py ends here



"""Provides an enhanced binding class."""

##############################################################################
# Textual imports.
from textual.binding import Binding


##############################################################################
class HelpfulBinding(Binding):
    """A binding that should show in the help screen.

    In many cases a binding will be associated with a
    [`Command`][textual_enhanced.commands.Command] and will show up in the
    [help screen][textual_enhanced.dialogs.HelpScreen] anyway. But sometimes
    there will be bindings that are particular to a widget that I want
    highlighted, that I want called out in the help screen. On the other
    hand I don't want *all* bindings to show in the help screen as that
    would end up being a cluttered and unhelpful mess.

    This class lets a binding be marked as helpful, destined for the help
    screen.
    """

    @property
    def most_helpful_description(self) -> str:
        """The most helpful description possible."""
        return self.tooltip or self.description


### binding.py ends here



"""A version of Textual's OptionList with some more navigation options."""

##############################################################################
# Python imports.
from types import TracebackType

##############################################################################
# Textual imports.
from textual.binding import Binding
from textual.geometry import Size
from textual.widgets import OptionList
from textual.widgets.option_list import OptionDoesNotExist

##############################################################################
# Typing extension imports.
from typing_extensions import Self


##############################################################################
class PreservedHighlight:
    """Context manager class to preserve an `OptionList` location.

    If the highlighted option has an ID, an attempt will be made to get back
    to that option; otherwise we return to the option in the same location.
    """

    def __init__(self, option_list: OptionList) -> None:
        """Initialise the object.

        Args:
            option_list: The `OptionList` to preserve the location for.
        """
        self._option_list = option_list
        """The option list we're preserving the location for."""
        self._highlighted = option_list.highlighted
        """The highlight that we should try to go back to."""
        self._option_id = (
            option_list.get_option_at_index(self._highlighted).id
            if self._highlighted is not None
            else None
        )
        """The ID of the option to try and get back to, or `None`."""

    def __enter__(self) -> Self:
        """Handle entry to the context."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        """Handle exit from the context."""
        del exc_type, exc_val, exc_traceback
        # Attempt to get back to the same option, or an option in a similar
        # location.
        try:
            self._option_list.highlighted = (
                self._highlighted
                if self._option_id is None
                else self._option_list.get_option_index(self._option_id)
            )
        except OptionDoesNotExist:
            self._option_list.highlighted = self._highlighted
        # If we still haven't landed anywhere and there are options, select
        # the first one.
        if self._option_list.highlighted is None and self._option_list.option_count:
            self._option_list.highlighted = 0


##############################################################################
class EnhancedOptionList(OptionList):
    """The Textual `OptionList` with more features.

    Key changes are:

    - By default some `vim`/`less`-user-friendly key bindings are added for
      navigation.
    - [`get_content_width`][textual_enhanced.widgets.EnhancedOptionList]
      works around a known Textual bug.
    - Adds
      [`preserved_highlight`][textual_enhanced.widgets.EnhancedOptionList.preserved_highlight].
    """

    DEFAULT_CSS = """
    EnhancedOptionList {
        &:focus {
            background-tint: initial;
        }
    }
    """

    BINDINGS = [
        Binding("j, right", "cursor_down", show=False),
        Binding("k, left", "cursor_up", show=False),
        Binding("<", "first", show=False),
        Binding(">", "last", show=False),
        Binding("space", "page_down", show=False),
    ]

    @property
    def preserved_highlight(self) -> PreservedHighlight:
        """Provides a context that preserves the highlight location.

        The rule for how this works is:

        - If the highlighted option has an ID, an attempt will be made to
          get back to that option;
        - Failing the above an attempt will be made to return to the same
          *location* in the list.
        - If neither approach can be taken, and if there `OptionList` has
          any content, the first option will be highlighted.

        The benefit of this is that the content of an `OptionList` can be
        cleared and built up again and, especially if your options have IDs,
        the user will be no wiser that such drastic measures were taken.

        Example:
            ```python
            with self.query_one(EnhancedOptionList).preserved_highlight:
                ...do things that modify the `OptionList`
            ```
        """
        return PreservedHighlight(self)

    def get_content_width(self, container: Size, viewport: Size) -> int:
        """Workaround for [textual#5489](https://github.com/Textualize/textual/issues/5489).

        Note:
            [textual#5489](https://github.com/Textualize/textual/issues/5489)
            has been fixed in later versions of Textual, but textual >=2.0
            still has some problems so I'm holding off using it for now.
        """
        return (
            super().get_content_width(container, viewport) if self.option_count else 0
        )


### option_list.py ends here



"""Provides a base command-oriented command palette provider class."""

##############################################################################
# Python imports.
from abc import abstractmethod
from collections.abc import Iterator
from functools import partial
from typing import NamedTuple, TypeAlias

##############################################################################
# Rich imports.
from rich.text import Text

##############################################################################
# Textual imports.
from textual.command import DiscoveryHit, Hit, Hits, Provider
from textual.content import Content
from textual.message import Message
from textual.types import IgnoreReturnCallbackType

##############################################################################
# Local imports.
from .bindings import primary_key_for
from .command import Command


##############################################################################
class CommandHit(NamedTuple):
    """A command hit for use in building a command palette hit."""

    command: str
    """The command."""
    description: str
    """The description of the command."""
    message: Message
    """The message to emit when the command is chosen."""


##############################################################################
CommandHits: TypeAlias = Iterator[CommandHit | Command]
"""The result of looking for commands to make into hits."""


##############################################################################
class CommandsProvider(Provider):
    """A base class for command-message-oriented command palette commands."""

    @classmethod
    def prompt(cls) -> str:
        """The prompt for the command provider."""
        return ""

    def maybe(self, command: type[Command]) -> CommandHits:
        """Yield a command if it's applicable.

        Args:
            command: The type of the command to maybe yield.

        Yields:
            The command if it can be used right now.

        This method takes the command, looks at its `action_name` and uses
        Textual's `check_action` to see if the action can be performed right
        now. If it can it will `yield` an instance of the command, otherwise
        it does nothing.
        """
        if self.screen.check_action(command.action_name(), ()):
            yield command()

    @abstractmethod
    def commands(self) -> CommandHits:
        """Provide the command data for the command palette.

        Yields:
            A tuple of the command, the command description and a command message.
        """
        raise NotImplementedError

    @property
    def _commands(self) -> Iterator[CommandHit]:
        """The commands available for the palette."""
        return (
            CommandHit(command.context_command, command.context_tooltip, command)
            if isinstance(command, Command)
            else command
            for command in self.commands()
        )

    def _maybe_add_binding(
        self, message: Command | Message, text: str | Text | Content
    ) -> Text | Content:
        """Maybe add binding details to some text.

        Args:
            message: The command message to maybe get the binding for.
            text: The text to add the binding details to.

        Returns:
            The resulting text.
        """
        if isinstance(text, str):
            text = Text(text)
        if not isinstance(message, Command) or not message.has_binding:
            return text
        key = Text(
            f"[{primary_key_for(self.screen, message.__class__)}]",
            style=(self.app.current_theme.accent if self.app.current_theme else None)
            or "dim",
        )
        # There's a breaking change between Textual 1.0 and 2.0
        # regarding how commands in the command palette are handled. I
        # could report it but I know the experience will be exhausting
        # and fruitless; so let's just handle it here.
        #
        # Eventually, once I'm happy that Textual 2.x is stable enough
        # to use, I'll tidy this up.
        if isinstance(text, Text):
            return text.append_text(Text(" ")).append_text(key)
        return Content.assemble(text, " ", Content.from_rich_text(key))

    def _perform(self, message: Command | Message) -> IgnoreReturnCallbackType:
        """Create the call to perform a command.

        Args:
            message: The message to perform.

        Returns:
            The call to perform the command.
        """
        if isinstance(message, Command):
            return partial(self.screen.run_action, message.action_name())
        return partial(self.screen.post_message, message)

    async def discover(self) -> Hits:
        """Handle a request to discover commands.

        Yields:
            Command discovery hits for the command palette.
        """
        for command, description, message in self._commands:
            yield DiscoveryHit(
                self._maybe_add_binding(message, command),
                self._perform(message),
                help=description,
            )

    async def search(self, query: str) -> Hits:
        """Handle a request to search for commands that match the query.

        Args:
            query: The query from the user.

        Yields:
            Command hits for the command palette.
        """
        matcher = self.matcher(query)
        for command, description, message in self._commands:
            if match := matcher.match(command):
                yield Hit(
                    match,
                    self._maybe_add_binding(message, matcher.highlight(command)),
                    self._perform(message),
                    help=description,
                )


### provider.py ends here



"""Provides a function for adding a key name to a label."""

##############################################################################
# Python imports.
from typing import Any

##############################################################################
# Textual imports.
from textual.app import App
from textual.widget import Widget


##############################################################################
def add_key(label: str, key: str, context: App[Any] | Widget | None = None) -> str:
    """Add a key name to a label.

    Args:
        label: The label to add the key to.
        key: The display name of the key to add.
        context: The widget, screen or application.

    Returns:
        The label with a display of the key added.

    Notes:
        By default the key label will use the current theme's accent colour,
        if that can't be derived from the context them `dim` will be used.
    """
    if isinstance(context, Widget):
        context = context.app
    key_colour = (
        "dim"
        if context is None or context.current_theme is None
        else context.current_theme.accent
    )
    return f"{label} [{key_colour}]\\[{key}][/]"


### add_key_name.py ends here



"""A widget for viewing text."""

##############################################################################
# Textual imports.
from textual.widgets import TextArea


##############################################################################
class TextViewer(TextArea):
    """A widget for viewing text."""

    DEFAULT_CSS = """
    TextViewer {
        background: transparent;
        border: none;
        &:focus {
            border: none;
        }
        & > .text-area--cursor-line {
           background: transparent;
        }
    }
    """

    BINDINGS = [
        ("<", "cursor_line_start"),
        (">", "cursor_line_end"),
        ("c, C", "copy"),
    ]

    def __init__(
        self,
        text: str = "",
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        """Initialise the object.

        Args:
            text: The text to view.
            name: The name of the TextViewer.
            id: The ID of the TextViewer in the DOM.
            classes: The CSS classes of the TextViewer.
            disabled: Whether the TextViewer is disabled or not.
        """
        super().__init__(
            text,
            read_only=True,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )

    def action_copy(self) -> None:
        """Action for copying text to the clipboard."""
        if for_clipboard := self.selected_text:
            self.notify("Selected text copied to the clipboard.")
        else:
            for_clipboard = self.text
            self.notify("All text copied to the clipboard.")
        self.app.copy_to_clipboard(for_clipboard)

    def action_cursor_line_start(self, select: bool = False) -> None:
        """Add a slightly smarter use of going 'home'.

        The rule with this version of going home is that if the cursor isn't
        at the start of the line, then the normal
        [`TextArea`][textual.widgets.TextArea] rules apply, otherwise if the
        cursor *is* at the start of the line the cursor will go to the start
        of the document.
        """
        if self.cursor_at_start_of_line:
            self.move_cursor(self.document.start, select=select)
        else:
            super().action_cursor_line_start(select)

    def action_cursor_line_end(self, select: bool = False) -> None:
        """Add a slightly smarter use of going 'end'.

        The rule with this version of going to the end is that if the cursor
        isn't at the end of the line, then the normal
        [`TextArea`][textual.widgets.TextArea] rules apply, otherwise if the
        cursor *is* at the end of the line the cursor will go to the end of
        the document.
        """
        if self.cursor_at_end_of_line:
            self.move_cursor(self.document.end, select=select)
        else:
            super().action_cursor_line_end(select)


### text_viewer.py ends here



"""Provides enhanced container classes."""

##############################################################################
# Python imports.
from typing import Literal

##############################################################################
# Textual imports.
from textual.containers import VerticalScroll

##############################################################################
# Local imports.
from .binding import HelpfulBinding


##############################################################################
class EnhancedVerticalScroll(VerticalScroll):
    """A vertical scroll container with some enhancements.

    The main purpose of this class is to add some extra bindings for
    scrolling, and to add a couple more scrolling actions.
    """

    BINDINGS = [
        HelpfulBinding("j, e, enter", "scroll_down", tooltip="Scroll down one line"),
        HelpfulBinding("k, y", "scroll_up", tooltip="Scroll up one line"),
        HelpfulBinding("f, space, z", "page_down", tooltip="Scroll down one page"),
        HelpfulBinding("b, w", "page_up", tooltip="Scroll up one page"),
        HelpfulBinding(
            "shift+pageup, u", "scroll_half_page(-1)", tooltip="Scroll up half a page"
        ),
        HelpfulBinding(
            "shift+pagedown, d",
            "scroll_half_page(1)",
            tooltip="Scroll down half a page",
        ),
        HelpfulBinding("g, <, p, %", "scroll_home", tooltip="Scroll to the top"),
        HelpfulBinding("G, >", "scroll_end", tooltip="Scroll to the bottom"),
    ]
    """Additional movement bindings that might help vim/less users."""

    def action_scroll_half_page(self, direction: Literal[-1, 1]) -> None:
        """Scroll the view half a page in the given direction.

        Args:
            direction: The direction to scroll in.
        """
        self.scroll_relative(y=(self.size.height // 2) * direction)


### containers.py ends here



"""Command classes and a provider for common commands."""

##############################################################################
# Local imports.
from .command import Command
from .provider import CommandHits, CommandsProvider


##############################################################################
class Help(Command):
    """Show help for and information about the application"""

    BINDING_KEY = "f1, ?"
    SHOW_IN_FOOTER = True


##############################################################################
class ChangeTheme(Command):
    """Change the application's theme"""

    BINDING_KEY = "f9"


##############################################################################
class Quit(Command):
    """Quit the application"""

    BINDING_KEY = "f10, ctrl+q"
    SHOW_IN_FOOTER = True
    ACTION = "app.quit"


##############################################################################
class CommonCommands(CommandsProvider):
    """Provides some common top-level commands for the application."""

    def commands(self) -> CommandHits:
        """Provide the main application commands for the command palette.

        Yields:
            The commands for the command palette.
        """
        yield Help()
        yield ChangeTheme()
        yield Quit()


### common.py ends here



"""A Textual screen, with tweaks."""

##############################################################################
# Python imports.
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Generic

##############################################################################
# Textual imports.
from textual.command import CommandPalette
from textual.css.query import QueryType
from textual.screen import Screen, ScreenResultType

##############################################################################
# Local imports.
from .commands import CommandsProvider
from .dialogs import HelpScreen


##############################################################################
class EnhancedScreen(Generic[ScreenResultType], Screen[ScreenResultType]):
    """A Textual screen with some extras."""

    @contextmanager
    def busy_looking(self, selector: str | type[QueryType]) -> Iterator[None]:
        """Provides a context that makes a widget look busy.

        A simple helper that turns this:

        ```python
        self.query_one(Display).loading = True
        self.do_something_that_takes_a_moment()
        self.query_one(Display).loading = False
        ```

        into this:

        ```python
        with self.busy_looking(Display):
            self.do_something_that_takes_a_moment()
        ```
        """
        (busy_widget := self.query_one(selector)).loading = True
        try:
            yield
        finally:
            busy_widget.loading = False

    def show_palette(self, provider: type[CommandsProvider]) -> None:
        """Show a particular command palette.

        Args:
            provider: The commands provider for the palette.
        """
        self.app.push_screen(
            CommandPalette(
                providers=(provider,),
                placeholder=provider.prompt(),
            )
        )

    def action_help_command(self) -> None:
        """Show the help screen.

        Rather than use Textual's own help facility, this shows [my own help
        screen][textual_enhanced.dialogs.HelpScreen].
        """
        self.app.push_screen(HelpScreen(self))

    def action_change_theme_command(self) -> None:
        """Show the Textual theme picker command palette."""
        self.app.search_themes()


### screen.py ends here



"""Tweaked base application class."""

##############################################################################
# Python imports.
from typing import Generic

##############################################################################
# Textual imports.
from textual.app import App, ReturnType
from textual.binding import Binding


##############################################################################
class EnhancedApp(Generic[ReturnType], App[ReturnType]):
    """The Textual [App class][textual.app.App]  with some styling tweaks.

    `EnhancedApp` adds no code changes, but it does implement a number of
    global styles that make a Textual app look just how I like. It also adds
    some extra default bindings for calling the command palette.
    """

    CSS = """
    CommandPalette > Vertical {
        width: 75%; /* Full-width command palette looks kinda unfinished. Fix that. */
        background: $panel;
        #--input {
            border-top: hkey $border;
        }
        OptionList{
            scrollbar-background: $panel;
            scrollbar-background-hover: $panel;
            scrollbar-background-active: $panel;
            border-bottom: hkey $border;
        }
        SearchIcon {
            display: none;
        }
    }

    /* Make the LoadingIndicator look less like it was just slapped on. */
    LoadingIndicator {
        background: transparent;
    }

    /* Remove cruft from the Header. */
    Header {
        /* I have zero use for the header icon or the clock. */
        HeaderIcon, HeaderClockSpace {
            display: none;
        }

        /* Ditto the tall version of the header. Nuke that. */
        &.-tall {
            height: 1 !important;
        }
    }

    /* General style tweaks that affect all widgets. */
    * {
        /* Let's make scrollbars a wee bit thinner. */
        scrollbar-size-vertical: 1;
    }
    """

    BINDINGS = [
        Binding(
            "ctrl+p, super+x, alt+x, :",
            "command_palette",
            "Commands",
            show=False,
            tooltip="Show the command palette",
        ),
    ]


### app.py ends here



"""Provides helper code for dealing with bindings."""

##############################################################################
# Python imports.
from collections.abc import Iterator

##############################################################################
# Textual imports.
from textual.app import App
from textual.binding import Binding
from textual.dom import DOMNode
from textual.screen import Screen

##############################################################################
# Local imports.
from .command import Command


##############################################################################
def all_keys_for(node: DOMNode, source: type[Command] | Binding) -> Iterator[str]:
    """Get all the keys for the given command or binding.

    Args:
        node: The node we're working from.
        source: The command or binding to get the keys for.

    Yields:
        The display names of all the keys for the command/binding.
    """
    for binding in (
        [source]
        if isinstance(source, Binding)
        else [
            binding.binding
            for binding in (
                node if isinstance(node, (App, Screen)) else node.screen
            ).active_bindings.values()
            if binding.binding.id == source.__name__
        ]
        or [source.binding()]
    ):
        for key in binding.key.split(","):
            yield node.app.get_key_display(Binding(key.strip(), ""))


##############################################################################
def primary_key_for(node: DOMNode, source: type[Command] | Binding) -> str:
    """Get the primary key for the given command or binding.

    Args:
        node: The node we're working from.
        source: The command or binding to get the keys for.

    Returns:
        The display name of the primary key for the command/binding.
    """
    return next(all_keys_for(node, source))


### bindings.py ends here



"""Provides a history tracking class."""

##############################################################################
# Python imports.
from collections import deque
from collections.abc import Iterator, Sequence
from typing import Generic, TypeVar

##############################################################################
# Textual imports.
from textual.geometry import clamp

##############################################################################
# Typing extensions imports.
from typing_extensions import Self

##############################################################################
HistoryItem = TypeVar("HistoryItem")
"""The type of an item in history."""


##############################################################################
class History(Generic[HistoryItem]):
    """A class for handling and tracking history."""

    def __init__(
        self, history: Sequence[HistoryItem] | None = None, max_length: int = 500
    ) -> None:
        """Initialise the history object.

        Args:
            history: Set to the given history.
            max_length: Optional maximum length for the history.
        """
        self._history: deque[HistoryItem] = deque(history or [], maxlen=max_length)
        """The history."""
        self._current = max(len(self._history) - 1, 0)
        """The current location within the history."""

    @property
    def current_location(self) -> int | None:
        """The current integer location in the history.

        If there is no valid location the value is `None`.
        """
        try:
            _ = self._history[self._current]
        except IndexError:
            return None
        return self._current

    @property
    def current_item(self) -> HistoryItem | None:
        """The current item in the history.

        If there is no current item in the history the value is `None`.
        """
        try:
            return self._history[self._current]
        except IndexError:
            return None

    @property
    def can_go_backward(self) -> bool:
        """Can history go backward?"""
        return bool(self._current)

    def backward(self) -> bool:
        """Go backward through the history.

        Returns:
            `True` if we moved through history, `False` if not.
        """
        if self.can_go_backward:
            self._current -= 1
            return True
        return False

    @property
    def can_go_forward(self) -> bool:
        """Can history go forward?"""
        return self._current < len(self._history) - 1

    def forward(self) -> bool:
        """Go forward through the history.

        Returns:
            `True` if we moved through history, `False` if not.
        """
        if self.can_go_forward:
            self._current += 1
            return True
        return False

    def goto(self, location: int) -> Self:
        """Jump to a specific location within history."""
        self._current = clamp(location, 0, len(self._history) - 1)
        return self

    def goto_end(self) -> Self:
        """Go to the end of the history."""
        self.goto(len(self) - 1)
        return self

    def add(self, item: HistoryItem) -> Self:
        """Add an item to the history.

        Args:
            item: The item to add.

        Returns:
            Self.
        """
        self._history.append(item)
        return self.goto_end()

    def __len__(self) -> int:
        """The length of the history."""
        return len(self._history)

    def __iter__(self) -> Iterator[HistoryItem]:
        """Support iterating through the history."""
        return iter(self._history)

    def __delitem__(self, index: int) -> None:
        """Delete an item in history."""
        del self._history[index]


### history.py ends here



"""A simple demo app of some of the enhancements."""

##############################################################################
# Python imports.
from dataclasses import dataclass

##############################################################################
# Textual imports.
from textual import on, work
from textual.app import ComposeResult, RenderResult
from textual.containers import Horizontal
from textual.message import Message
from textual.widgets import Button, Footer, Header

##############################################################################
# Textual Enhanced imports.
from textual_enhanced import __version__
from textual_enhanced.app import EnhancedApp
from textual_enhanced.binding import HelpfulBinding
from textual_enhanced.commands import (
    ChangeTheme,
    Command,
    CommandHit,
    CommandHits,
    CommandsProvider,
    CommonCommands,
    Help,
    Quit,
)
from textual_enhanced.dialogs import Confirm, ModalInput
from textual_enhanced.screen import EnhancedScreen


##############################################################################
@dataclass
class ShowNumber(Message):
    number: int


##############################################################################
class NumberProvider(CommandsProvider):
    def commands(self) -> CommandHits:
        for n in range(500):
            yield CommandHit(
                f"This is the rather special and unique number {n}",
                f"This is some help about the number {n}, just in case you needed it",
                ShowNumber(n),
            )


##############################################################################
class SayOne(Command):
    ACTION = "say('One')"


##############################################################################
class SayTwo(Command):
    ACTION = "say('Two')"


##############################################################################
class OtherCommands(CommandsProvider):
    def commands(self) -> CommandHits:
        yield SayOne()
        yield SayTwo()


##############################################################################
class HelpfulButton(Button):
    BINDINGS = [
        HelpfulBinding("ctrl+o", "gndn", description="This does nothing useful")
    ]


##############################################################################
class Ruler(Horizontal):
    DEFAULT_CSS = """
    Ruler {
        width: 1fr;
        height: 1;
        text-align: center;
    }
    """

    def render(self) -> RenderResult:
        return "----- | -----"


##############################################################################
class Main(EnhancedScreen[None]):
    TITLE = "Title"
    SUB_TITLE = "Title"
    COMMAND_MESSAGES = (Help, ChangeTheme, Quit)
    COMMANDS = {CommonCommands, OtherCommands}
    BINDINGS = Command.bindings(
        *COMMAND_MESSAGES,
        HelpfulBinding(
            "ctrl+y, ctrl+i",
            "gndn",
            show=False,
            description="This is the description",
            tooltip="This is the tooltip",
        ),
        HelpfulBinding(
            "vertical_line, ctrl+t, ctrl+k",
            "gndn",
            description="This is the description",
        ),
    )
    HELP = "Here's some really long text to scroll: {}".format(
        "\n".join([f"{n}. This is a filler line" for n in range(100)])
    )

    def compose(self) -> ComposeResult:
        yield Header()
        yield Ruler()
        yield HelpfulButton("Quick input", id="input")
        yield Button("Another input", id="input_with_default")
        yield Button("Yes or no?", id="confirm")
        yield Button("Pick a number", id="number")
        yield Footer()

    @on(Button.Pressed, "#input, #input_with_default")
    @work
    async def input_action(self, message: Button.Pressed) -> None:
        if text := await self.app.push_screen_wait(
            ModalInput(placeholder="Enter some text here")
            if message.button.id == "input"
            else ModalInput(
                placeholder="This has an initial value", initial="Testing..."
            )
        ):
            self.notify(f"Entered '{text}")

    @on(Button.Pressed, "#confirm")
    @work
    async def confirm_action(self) -> None:
        self.notify(
            "YES!"
            if await self.app.push_screen_wait(
                Confirm(
                    "Well?", "So, what's the decision? Are we going with yes or no?"
                )
            )
            else "No!"
        )

    @on(Button.Pressed, "#number")
    def pick_a_number(self) -> None:
        self.show_palette(NumberProvider)

    @on(ShowNumber)
    def show_the_number(self, number: ShowNumber) -> None:
        self.notify(f"You picked {number.number}")

    def action_say(self, text: str) -> None:
        self.notify(text)


##############################################################################
class DemoApp(EnhancedApp[None]):
    """A little demo app."""

    HELP_TITLE = f"textual-enhanced v{__version__}"
    HELP_ABOUT = "A library of mildly-opinionated enhancements to Textual."
    HELP_LICENSE = """MIT License

    Copyright (c) 2025 Dave Pearson <davep@davep.org>

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
    IN THE SOFTWARE.
    """

    COMMANDS = set()

    def get_default_screen(self) -> Main:
        return Main()

    def on_mount(self) -> None:
        self.update_keymap({"Help": "f7,f1,h,question_mark"})


##############################################################################
if __name__ == "__main__":
    DemoApp().run()

### __main__.py ends here



"""A library of opinionated enhancements for Textual."""

##############################################################################
# Python imports.
from importlib.metadata import version

######################################################################
# Main app information.
__author__ = "Dave Pearson"
__copyright__ = "Copyright 2025, Dave Pearson"
__credits__ = ["Dave Pearson"]
__maintainer__ = "Dave Pearson"
__email__ = "davep@davep.org"
__version__ = version("textual_enhanced")
__licence__ = "MIT"

### __init__.py ends here



"""Provides code related to 'commands' in a Textual application."""

##############################################################################
# Local imports.
from .command import Command
from .common import ChangeTheme, CommonCommands, Help, Quit
from .provider import CommandHit, CommandHits, CommandsProvider

##############################################################################
# Exports.
__all__ = [
    "ChangeTheme",
    "Command",
    "CommandHit",
    "CommandHits",
    "CommandsProvider",
    "CommonCommands",
    "Help",
    "Quit",
]

### __init__.py ends here



"""Provides a simple help screen."""

##############################################################################
# Python imports.
from inspect import cleandoc
from operator import attrgetter, methodcaller
from typing import Any
from webbrowser import open as open_url

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.containers import Center, Vertical
from textual.dom import DOMNode
from textual.screen import ModalScreen, Screen
from textual.widgets import Button, Markdown

##############################################################################
# Textual enhanced imports.
from ..binding import HelpfulBinding
from ..commands.bindings import all_keys_for
from ..containers import EnhancedVerticalScroll
from ..tools import add_key


##############################################################################
class HelpScreen(ModalScreen[None]):
    """The help screen."""

    CSS = """
    HelpScreen {
        align: center middle;

        &> Vertical {
            width: 75%;
            height: 90%;
            background: $panel;
            border: solid $border;
        }

        Markdown, MarkdownTable {
            padding: 0 1 0 1;
            background: transparent;
        }

        MarkdownH1 {
            padding: 1 0 1 0;
            background: $foreground 10%;
        }

        EnhancedVerticalScroll {
            scrollbar-gutter: stable;
            scrollbar-background: $panel;
            scrollbar-background-hover: $panel;
            scrollbar-background-active: $panel;
        }

        Center {
            height: auto;
            width: 100%;
            border-top: solid $border;
        }
    }
    """

    BINDINGS = [("escape, f1", "close")]

    def __init__(self, help_for: Screen[Any] | None = None) -> None:
        """Initialise the help screen.

        Args:
            help_for: The screen to show the help for.
        """
        super().__init__()
        if help_for is None:
            help_for = self.app.screen
        self._context_help = ""
        for node in (
            help_for.focused if help_for.focused is not None else help_for
        ).ancestors_with_self:
            if node.HELP is not None:
                self._context_help += f"\n\n{cleandoc(node.HELP)}"
            self._context_help += self.input_help(node)

    @staticmethod
    def _table_safe(text: str) -> str:
        """Escape some text for use in a table.

        Args:
            text: The text to make safe.

        Returns:
            The text made safe for use in a table.
        """
        return text.replace("|", r"\|")

    def input_help(self, node: DOMNode) -> str:
        """Build help from the bindings and commands provided by a DOM node.

        Args:
            node: The node that might provide commands.

        Returns:
            The help text.
        """
        helpful_bindings = [
            binding
            for binding in getattr(node, "BINDINGS", [])
            if isinstance(binding, HelpfulBinding)
        ]
        commands = getattr(node, "COMMAND_MESSAGES", [])
        if not any((helpful_bindings, commands)):
            return ""
        keys = f"{'| Command ' if commands else ''}| Key | Description |\n{'| - ' if commands else ''}| - | - |\n"
        for binding in sorted(
            helpful_bindings, key=attrgetter("most_helpful_description")
        ):
            keys += (
                f"{'| ' if commands else ''}| "
                f"{self._table_safe(', '.join(all_keys_for(node, binding)))} "
                f"| {binding.most_helpful_description} |\n"
            )
        for command in sorted(commands, key=methodcaller("command")):
            keys += (
                f"| {command.command()} "
                f"| {self._table_safe(', '.join(all_keys_for(node, command)))} | {command.tooltip()} |\n"
            )
        return f"\n\n{keys}"

    @property
    def _template(self) -> str:
        """The template for the help text."""
        template = ""
        if title := getattr(self.app, "HELP_TITLE", ""):
            template += f"# {cleandoc(title)}\n\n"
        template += "{context_help}\n\n"
        if about := getattr(self.app, "HELP_ABOUT", ""):
            template += f"## About\n\n{cleandoc(about)}\n\n"
        if license := getattr(self.app, "HELP_LICENSE", ""):
            template += f"## License\n\n{cleandoc(license)}\n"
        return template

    def compose(self) -> ComposeResult:
        """Compose the layout of the help screen."""
        with Vertical() as help_screen:
            help_screen.border_title = "Help"
            with EnhancedVerticalScroll():
                yield Markdown(self._template.format(context_help=self._context_help))
            with Center():
                yield Button(add_key("Okay", "Esc", self))

    @on(Button.Pressed)
    def action_close(self) -> None:
        """Close the help screen."""
        self.dismiss(None)

    @on(Markdown.LinkClicked)
    def visit(self, event: Markdown.LinkClicked) -> None:
        """Visit any link clicked in the help."""
        open_url(event.href)


### help.py ends here



"""A modal screen for quickly getting an input value."""

##############################################################################
# Textual imports.
from textual import on
from textual.app import ComposeResult
from textual.screen import ModalScreen
from textual.widgets import Input


##############################################################################
class ModalInput(ModalScreen[str | None]):
    """A modal screen to get input from the user."""

    CSS = """
    ModalInput {
        align: center middle;

        Input, Input:focus {
            border: round $border;
            width: 60%;
            padding: 1;
            height: auto;
        }
    }
    """

    BINDINGS = [("escape", "escape")]

    def __init__(
        self,
        placeholder: str | None = None,
        initial: str = "",
        classes: str | None = None,
    ) -> None:
        """Initialise the object.

        Args:
            placeholder: The placeholder text to use.
            initial: The initial value for the input.
            classes: The CSS classes of the modal input.
        """
        super().__init__(classes=classes)
        self._placeholder = placeholder or ""
        """The placeholder to use for the input."""
        self._initial = initial
        """The initial value for the input."""

    def compose(self) -> ComposeResult:
        """Compose the input dialog."""
        yield Input(self._initial, placeholder=self._placeholder)

    @on(Input.Submitted)
    def accept_input(self) -> None:
        """Accept the input."""
        self.dismiss(self.query_one(Input).value.strip())

    def action_escape(self) -> None:
        """Escape out without getting the input."""
        self.dismiss(None)


### modal_input.py ends here
